#include "Artboard.h"
#include "Layer.h"
#include "Utils.h"
#include "import-font.h"
#include "ofFileUtils.h"
#include "ofMain.h"
#include "ofApp.h"
#include "ofUtils.h"
#include "ofxProfiler.h"
#include "ofWindowSettings.h"
#include <string>

//========================================================================
// MAIN APP
// shared_ptr to main app
// this is global, so we can call it in our binding functions

shared_ptr <VariableEditor::ofApp> app;

//========================================================================
// EMSCRIPTEN BINDING FUNCTIONS
// here we add functions to be called from javascript
#ifdef TARGET_EMSCRIPTEN

    string addExistingLayer(ofxVariableLab::Layer::Props props,
                            string layerID){
        if(props.fontPath == ""){ // TODO: cleaner handling
            ofxVariableLab::Layer::Props p;
            props.fontPath = p.fontPath;
        }
        return app->layerComposition.addLayer(props,
                                              layerID);
    }
    string addNewLayer(){
        ofxVariableLab::Layer::Props props;
        return app->layerComposition.addLayer(props);
    }
    void removeLayer(string layerID){
        app->layerComposition.removeLayer(layerID);
    }
    vector <string> debug_getLayers(){
        vector <string> out;
        auto & layersMap = app->layerComposition.getLayers();
        for(const auto & [k, v] : layersMap){
            out.push_back(k);
        }
        return out;
    }
    void debug_downloadFbo(string name){
        if(name == "artboard"){
            app->downloadFboAsImage("artboard" + ofGetTimestampString() + ".png", app->artboard.getFbo());
        }else if(name == "fbo"){
            app->downloadFboAsImage("fbo" + ofGetTimestampString() + ".png", app->fbo);
        }
    }
    ofxVariableLab::Layer::Props getProps(string layerID){
        return app->layerComposition.getLayer(layerID)->getProps();
    }
    bool setProps(ofxVariableLab::Layer::Props props,
                  string layerID){
        props.letterDelay /= app->timeScale;
        for(const auto & [k, v]: props.letterDelays){
            props.letterDelays[k] /= app->timeScale;
        }
        app->layerComposition.getLayer(layerID)->setProps(props);
        return true; // TODO: do not always return true maybe
    }
    VariableEditor::ArtboardProps getArtboardProps(){
        return app->artboard.getProps();
    }
    bool setArtboardProps(VariableEditor::ArtboardProps props){
        app->artboard.setProps(props);
        return true; // TODO: do not always return true maybe
    }
    ofxVariableLab::Layer::BoundingBox getBoundingBox(string layerID){
        return app->layerComposition.getLayer(layerID)->getBoundingBox();
    }
    void resetLetterDelay(){
        for(const auto & l: app->layerComposition.getLayers()){
            l.second->clearPropsBuffer();
        }
    }
    void setLayerOrder(vector <string> _layerOrder){
        vector <ofxVariableLab::LayerID> layerOrder;
        for(int i = 0; i < _layerOrder.size(); i++){
            ofxVariableLab::LayerID layer = _layerOrder[i];
            layerOrder.push_back(layer);
        }
        app->layerComposition.setLayerOrder(_layerOrder);
    }
    void setPlaying(bool playing){
        app->setPlaying(playing);
    }
// TODO: rename this to exportRendering or sth similar
    void setRendering(bool rendering){
        ofDirectory exportDir(app->settings.tmpExportDir + "/frames");
        if(!exportDir.exists()){
            exportDir.create(true);
        }
        if(rendering){
            app->currentFrame = 0;
            app->recordedFrameNames.clear();
            exportDir.listDir();
            for(ofFile file : exportDir.getFiles()){
                //file.removeFile(file.getAbsolutePath(), false);
                file.remove();
            }
            double timelineLenth_seconds = EM_ASM_DOUBLE({
            return window.tp.core.val(window.tp.sheet.sequence.pointer.length);
        });
            double fps = 30.0;
            app->guessedFrameCount = std::ceil((fps * timelineLenth_seconds) / app->timeScale);
            app->timelineLength_seconds = timelineLenth_seconds;
        }
        app->rendering = rendering;
    }

    bool getRendering(){
        return app->rendering;
    }

    void renderNextFrame(){
        app->renderNextFrame = true;
    }

    void setTimeScale(float timeScale){
        app->timeScale = timeScale;
    }

    std::string importProjectAsZip(std::string filepath, std::string outdir){
        ofDisableDataPath();
        ofDirectory dir(outdir);
        if(!dir.exists()){
            dir.createDirectory(outdir);
        }
        VariableEditor::UnZip unzip(filepath, outdir);
        dir.listDir();

        ofDirectory userFonts("/idbfs/fonts");
        ofDirectory userAudio("/idbfs/audio"); // possibly rename to media?

        nlohmann::json json;
        json["files"] = nlohmann::json::array();
        for(int i = 0; i < dir.size(); i++){
            ofFile file = dir.getFile(i);
            if(ofToLower(file.getAbsolutePath()).find("project.json") != std::string::npos){
                try{
                    std::ifstream project(file.getAbsolutePath());
                    project >> json["project"];
                }
                catch(nlohmann::json::exception & e){
                    std::cerr << "Module::importProjectAsZip " << e.what() << std::endl;
                }

            }
            if(ofToLower(file.getExtension()) == "ttf"
               || ofToLower(file.getExtension()) == "otf"){
                if(!file.moveTo(userFonts)){
                    std::cout << "Module::importProjectAsZip" << " - cannot move font " << file.getFileName();
                }
            }else if(ofToLower(file.getExtension()) == "mp3"
                     || ofToLower(file.getExtension()) == "m4a"
                     || ofToLower(file.getExtension()) == "wav"
                     || ofToLower(file.getExtension()) == "webm"
                     || ofToLower(file.getExtension()) == "aac"
                     || ofToLower(file.getExtension()) == "ogg"){
                if(!file.moveTo(userAudio)){
                    std::cout << "Module::importProjectAsZip" << " - cannot move audiofile " << file.getFileName();
                }
            }else{
                file.remove();
            }
            json["files"].push_back(dir.getPath(i));
        }
        EM_ASM({
        FS.syncfs(false, function(err){
        });
    });
        return json.dump();
    }

    void exportFramesAsZip(string projectName){
        app->downloadFramesAsZip(projectName);
    }

    void downloadProject(string projectName,
                         string projectJsonString){
        app->downloadProject(projectName,
                             projectJsonString);
    }

    std::vector <std::string> listAvailableFonts(){
        vector <std::string> out;
        ofEnableDataPath();
        ofDirectory rootDir(".");
        rootDir.listDir();
        {
            ofDirectory fontsDir("fonts");
            fontsDir.sort();
            fontsDir.allowExt("ttf");
            fontsDir.allowExt("otf");
            fontsDir.listDir();
            for(int i = 0; i < fontsDir.size(); i++){
                std::string fontPath = "data/" + fontsDir.getPath(i);
                out.push_back(fontPath);
            }
        }
        {
            ofDirectory fontsDir("free-fonts");
            fontsDir.sort();
            fontsDir.allowExt("ttf");
            fontsDir.allowExt("otf");
            fontsDir.listDir();
            for(int i = 0; i < fontsDir.size(); i++){
                std::string fontPath = "data/" + fontsDir.getPath(i);
                out.push_back(fontPath);
            }
        }
        ofDirectory idbfsDir("/idbfs");
        if(idbfsDir.exists()){
            ofDirectory userFonts("/idbfs/fonts");
            if(userFonts.exists()){
                userFonts.sort();
                userFonts.allowExt("ttf");
                userFonts.allowExt("otf");
                userFonts.listDir();
                for(int i = 0; i < userFonts.size(); i++){
                    std::string fontPath = userFonts.getPath(i);
                    ofStringReplace(fontPath, "/idbfs/", "idbfs/");
                    out.push_back(fontPath);
                }
            }
        }
        return out;
    }
    std::vector <ofxVariableLab::FontVariationAxis> listVariationAxes(std::string fontPath){
        std::vector <ofxVariableLab::FontVariationAxis> axes;
        ofxVariableLab::listFontVariationAxes(fontPath, axes);
        return axes;
    }

    void windowResized(int w, int h){
        app->windowResized(w, h);
    }

    void setUniform1f(string name, float value){
        cout << "VariableEditor::setUniform1f " << name << ": " << ofToString(value) << endl;
        app->layerComposition.setUniform1fv({name}, {value});
    }

    void setUniform1i(string name, int value){
        app->layerComposition.setUniform1iv({name}, {value});
    }

    void startProfiling(){
        #if OFX_PROFILER
            OFX_PROFILER_BEGIN_SESSION("profiling", "results.json");
        #else
            std::cout << "not compiled with internal profiling support" << endl;
            std::cout << "make sure to include ofxProfiler and define OFX_PROFILER=1" << endl;
        #endif
    }
    void endProfiling(){
        #if OFX_PROFILER
            OFX_PROFILER_END_SESSION();

            #ifdef TARGET_EMSCRIPTEN
                std::ifstream f("results.json");
                nlohmann::json json = nlohmann::json::parse(f);
                app->downloadJson("results.json", json);
            #endif
        #else
            std::cout << "not compiled with internal profiling support" << endl;
            std::cout << "make sure to include ofxProfiler and define OFX_PROFILER=1" << endl;
        #endif
    }

//========================================================================
// EMSCRIPTEN BINDING BINDINGS
// here we describe classes and functions to bind them to javascript

    EMSCRIPTEN_BINDINGS(name_is_irrelevant){
        emscripten::value_object <VariableEditor::ArtboardProps>("ArtboardProps")
        .field("backgroundColor", &VariableEditor::ArtboardProps::backgroundColor)
        .field("x", &VariableEditor::ArtboardProps::x)
        .field("y", &VariableEditor::ArtboardProps::y)
        .field("width", &VariableEditor::ArtboardProps::width)
        .field("height", &VariableEditor::ArtboardProps::height)
        .field("zoom", &VariableEditor::ArtboardProps::zoom)
        .field("pixelDensity", &VariableEditor::ArtboardProps::pixelDensity)
        ;
        emscripten::value_object <ofxVariableLab::Layer::Props>("Props")
        .field("x", &ofxVariableLab::Layer::Props::x)
        .field("y", &ofxVariableLab::Layer::Props::y)
        .field("width", &ofxVariableLab::Layer::Props::width)
        .field("height", &ofxVariableLab::Layer::Props::height)
        .field("rotation", &ofxVariableLab::Layer::Props::rotation)
        .field("fontSize_px", &ofxVariableLab::Layer::Props::fontSize_px)
        .field("letterSpacing", &ofxVariableLab::Layer::Props::letterSpacing)
        .field("lineHeight", &ofxVariableLab::Layer::Props::lineHeight)
        .field("textAlignment", &ofxVariableLab::Layer::Props::textAlignment)
        .field("color", &ofxVariableLab::Layer::Props::color)
        .field("mirror_x", &ofxVariableLab::Layer::Props::mirror_x)
        .field("mirror_x_distance", &ofxVariableLab::Layer::Props::mirror_x_distance)
        .field("mirror_y", &ofxVariableLab::Layer::Props::mirror_y)
        .field("mirror_y_distance", &ofxVariableLab::Layer::Props::mirror_y_distance)
        .field("mirror_xy", &ofxVariableLab::Layer::Props::mirror_xy)
        .field("letterDelay", &ofxVariableLab::Layer::Props::letterDelay)
        .field("transformOrigin", &ofxVariableLab::Layer::Props::transformOrigin)
        .field("text", &ofxVariableLab::Layer::Props::text)
        .field("fontPath", &ofxVariableLab::Layer::Props::fontPath)
        .field("fontVariations", &ofxVariableLab::Layer::Props::fontVariations)
        .field("letterDelays", &ofxVariableLab::Layer::Props::letterDelays)
        ;
        emscripten::value_object <ofxVariableLab::FontVariationAxis>("FontVariationAxis")
        .field("name", &ofxVariableLab::FontVariationAxis::name)
        .field("minValue", &ofxVariableLab::FontVariationAxis::minValue)
        .field("maxValue", &ofxVariableLab::FontVariationAxis::maxValue)
        .field("defaultValue", &ofxVariableLab::FontVariationAxis::defaultValue)
        ;
        emscripten::value_object <ofxVariableLab::Layer::BoundingBox>("BoundingBox")
        .field("x", &ofxVariableLab::Layer::BoundingBox::x)
        .field("y", &ofxVariableLab::Layer::BoundingBox::y)
        .field("w", &ofxVariableLab::Layer::BoundingBox::w)
        .field("h", &ofxVariableLab::Layer::BoundingBox::h)
        ;
        // Register std::array<float, 4> because Props::color is interpreted as such
        emscripten::value_array <std::array <float, 4> >("array_float_4")
        .element(emscripten::index <0>())
        .element(emscripten::index <1>())
        .element(emscripten::index <2>())
        .element(emscripten::index <3>())
        ;
        emscripten::class_ <ofxVariableLab::FontVariation>("FontVariation")
        .constructor <>()
        .property("name", &ofxVariableLab::FontVariation::name)
        .property("value", &ofxVariableLab::FontVariation::value)
        ;
        emscripten::register_vector <ofxVariableLab::FontVariationAxis>("FontVariationAxes");
        emscripten::register_vector <ofxVariableLab::FontVariation>("FontVariations");
        emscripten::register_vector <std::string>("vector<string>");
        emscripten::register_map <std::string, float>("map<string, float>");

        emscripten::function("addExistingLayer", &addExistingLayer);
        emscripten::function("addNewLayer", &addNewLayer);
        emscripten::function("removeLayer", &removeLayer);
        emscripten::function("getProps", &getProps);
        emscripten::function("setProps", &setProps);
        //emscripten::function("listAvailableFontsAndAxes", &listAvailableFontsAndAxes);
        emscripten::function("listAvailableFonts", &listAvailableFonts);
        emscripten::function("listVariationAxes", &listVariationAxes);
        emscripten::function("getArtboardProps", &getArtboardProps);
        emscripten::function("setArtboardProps", &setArtboardProps);
        emscripten::function("getBoundingBox", &getBoundingBox);
        emscripten::function("resetLetterDelay", &resetLetterDelay);
        emscripten::function("setLayerOrder", &setLayerOrder);
        emscripten::function("setPlaying", &setPlaying);
        emscripten::function("setRendering", &setRendering);
        emscripten::function("getRendering", &getRendering);
        emscripten::function("setTimeScale", &setTimeScale);
        emscripten::function("windowResized", &windowResized);
        emscripten::function("exportFramesAsZip", &exportFramesAsZip);
        emscripten::function("importProjectAsZip", &importProjectAsZip, emscripten::allow_raw_pointers());
        emscripten::function("downloadProject", &downloadProject);
        // debug
        emscripten::function("setUniform1f", &setUniform1f);
        emscripten::function("setUniform1i", &setUniform1i);
        emscripten::function("startProfiling", &startProfiling);
        emscripten::function("endProfiling", &endProfiling);
        emscripten::function("debug_getLayers", &debug_getLayers);
        emscripten::function("debug_downloadFbo", &debug_downloadFbo);
    }
#endif

//========================================================================
// MAIN
// well, we have to start somewhere
int main(){
    #ifdef OF_TARGET_OPENGLES
        ofGLESWindowSettings settings;
        //settings.setSize(1920, 1080);
        settings.numSamples = 4;
        settings.glesVersion = 3;
    #else
        ofGLWindowSettings settings;
        settings.windowMode = OF_WINDOW;
        settings.setSize(1920, 1080);
        settings.setPosition(glm::vec2(1920, 0));
        settings.setGLVersion(3, 2);
    #endif
    ofCreateWindow(settings);

    app = make_shared <VariableEditor::ofApp>();
    // this kicks off the running of my app
    // can be OF_WINDOW or OF_FULLSCREEN
    // pass in width and height too:
    ofRunApp(app);
}
